@@ -1,30 +1,33 @@
 (ns puppetlabs.puppetdb.middleware
   "Ring middleware"
-  (:require [puppetlabs.i18n.core :as i18n]
+  (:require [puppetlabs.i18n.core :as i18n :refer [trs tru]]
             [puppetlabs.kitchensink.core :as kitchensink]
-            [puppetlabs.puppetdb.jdbc :as jdbc :refer [with-transacted-connection]]
+            [puppetlabs.puppetdb.jdbc :as jdbc]
             [puppetlabs.puppetdb.query-eng :as qe]
             [puppetlabs.puppetdb.utils.metrics :refer [multitime!]]
             [puppetlabs.puppetdb.http :as http]
+            [puppetlabs.puppetdb.schema :refer [defn-validated]]
+            [puppetlabs.ssl-utils.core :refer [get-cn-from-x509-certificate]]
             [ring.util.request :as request]
             [clojure.string :as str]
             [clojure.tools.logging :as log]
             [ring.middleware.params :refer [wrap-params]]
-            [puppetlabs.puppetdb.query.paging :as paging]
             [clojure.set :as set]
-            [pantomime.media :as media]
             [puppetlabs.puppetdb.metrics.core :as metrics]
-            [metrics.timers :refer [timer time!]]
+            [metrics.timers :refer [timer]]
             [metrics.meters :refer [meter mark!]]
             [metrics.histograms :refer [update!]]
-            [clojure.walk :refer [keywordize-keys]]
-            [puppetlabs.puppetdb.utils :as utils]
             [bidi.bidi :as bidi]
             [bidi.ring :as bring]
             [bidi.schema :as bidi-schema]
-            [puppetlabs.puppetdb.schema :as pls]
             [schema.core :as s]
-            [puppetlabs.puppetdb.command :as cmd]))
+            [puppetlabs.puppetdb.command :as cmd]
+            [puppetlabs.puppetdb.constants :as constants]
+            [puppetlabs.puppetdb.command.constants :as const])
+  (:import
+   (clojure.lang ExceptionInfo)
+   (java.net HttpURLConnection)
+   (java.sql SQLException)))
 
 (def handler-schema (s/=> s/Any {s/Any s/Any}))
 
@@ -35,28 +38,28 @@
   `<logger name=\"puppetlabs.puppetdb.middleware\" level=\"debug\"/>`"
   [app]
   (fn [req]
-    (log/debug (i18n/trs "Processing HTTP request to URI: ''{0}''" (:uri req)))
+    (log/debug (trs "Processing HTTP request to URI: ''{0}''" (:uri req)))
     (app req)))
 
-(defn build-whitelist-authorizer
+(defn build-allowlist-authorizer
   "Build a function that will authorize requests based on the supplied
-  certificate whitelist (see `cn-whitelist->authorizer` for more
+  certificate allowlist (see `cn-whitelist->authorizer` for more
   details). Returns :authorized if the request is allowed, otherwise a
   string describing the reason not."
-  [whitelist]
-  {:pre  [(string? whitelist)]
+  [allowlist]
+  {:pre  [(string? allowlist)]
    :post [(fn? %)]}
-  (let [allowed? (kitchensink/cn-whitelist->authorizer whitelist)]
+  (let [allowed? (kitchensink/cn-whitelist->authorizer allowlist)]
     (fn [{:keys [ssl-client-cn] :as req}]
       (when-not (allowed? req)
         (when ssl-client-cn
-          (log/warn (i18n/trs "{0} rejected by certificate whitelist {1}" ssl-client-cn whitelist)))
-        (http/denied-response (i18n/tru "The client certificate name {0} doesn't appear in the certificate whitelist. Is your master''s (or other PuppetDB client''s) certname listed in PuppetDB''s certificate-whitelist file?" ssl-client-cn)
-                         http/status-forbidden)))))
+          (log/warn (trs "{0} rejected by certificate allowlist {1}" ssl-client-cn allowlist)))
+        (http/denied-response (tru "The client certificate name {0} doesn't appear in the certificate allowlist. Is your master''s (or other PuppetDB client''s) certname listed in PuppetDB''s certificate-allowlist file?" ssl-client-cn)
+                              HttpURLConnection/HTTP_FORBIDDEN)))))
 
 (defn wrap-cert-authn
-  [app cert-whitelist]
-  (if-let [cert-authorize-fn (some-> cert-whitelist build-whitelist-authorizer)]
+  [app cert-allowlist]
+  (if-let [cert-authorize-fn (some-> cert-allowlist build-allowlist-authorizer)]
     (fn [req]
       (if-let [cert-auth-result (cert-authorize-fn req)]
         cert-auth-result
@@ -70,7 +73,7 @@
   the key's value is set to nil."
   [app]
   (fn [{:keys [ssl-client-cert] :as req}]
-    (let [cn  (some-> ssl-client-cert kitchensink/cn-for-cert)
+    (let [cn  (some-> ssl-client-cert get-cn-from-x509-certificate)
           req (assoc req :ssl-client-cn cn)]
       (app req))))
 
@@ -97,8 +100,67 @@
   (fn [req]
     (try
       (app req)
+      ;; This clause may become obsolete if everything it was handling
+      ;; should be (and eventually is) converted to bad-query-ex.
       (catch IllegalArgumentException e
-        (http/error-response e)))))
+        (http/error-response e))
+      (catch ExceptionInfo ex
+        (case (-> ex ex-data :kind)
+          :puppetlabs.puppetdb.query/invalid
+          (do ;; cf. produce-streaming-body
+            (cond
+              (seq (.getSuppressed ex)) (log/error ex)
+              (ex-cause ex) (log/info (ex-cause ex) (trs "Invalid query: {0}" (ex-message ex)))
+              :else (log/info (trs "Invalid query: {0}" (ex-message ex))))
+            (http/error-response ex))
+          (throw ex))))))
+
+(defn cause-finder
+  [ex]
+  (let [cause (.getCause ex)]
+    (if (nil? cause)
+      (.getMessage ex)
+      (recur cause))))
+
+(defn wrap-with-exception-handling
+  [app]
+  (fn [req]
+    (try
+      (app req)
+      (catch ExceptionInfo e
+        (if (= :puppetlabs.puppetdb.query/terminated (:kind (ex-data e)))
+          (http/error-response (tru "Query backend terminated")
+                               HttpURLConnection/HTTP_INTERNAL_ERROR)
+          (do
+            (log/error e)
+            (http/error-response (cause-finder e)
+                                 HttpURLConnection/HTTP_INTERNAL_ERROR))))
+      (catch SQLException e
+        (cond
+          (= (.getSQLState e) (jdbc/sql-state :admin-shutdown))
+          (do
+            (if (seq (.getSuppressed e))
+              (log/error e)
+              (log/error (.getMessage e)))
+            (http/error-response (.getMessage e)
+                                 HttpURLConnection/HTTP_UNAVAILABLE))
+          :else
+          (do
+            (log/error e)
+            (http/error-response (cause-finder e)
+                                 HttpURLConnection/HTTP_INTERNAL_ERROR))))
+      (catch Exception e
+        (log/error e)
+        (http/error-response (cause-finder e)
+                             HttpURLConnection/HTTP_INTERNAL_ERROR))
+      (catch AssertionError e
+        (log/error e)
+        (http/error-response (tru "An unexpected error occurred while processing the request")
+                             HttpURLConnection/HTTP_INTERNAL_ERROR))
+      (catch Throwable e
+        (log/error e)
+        (http/error-response (tru "An unexpected error occurred")
+                             HttpURLConnection/HTTP_INTERNAL_ERROR)))))
 
 (defn verify-accepts-content-type
   "Ring middleware that requires a request for the wrapped `app` to accept the
@@ -106,14 +168,29 @@
   Acceptable status is returned, with a message informing the client it must
   accept the content type."
   [app content-type]
-  {:pre (string? content-type)}
+  {:pre [(string? content-type)]}
   (fn [{:keys [headers] :as req}]
     (if (http/acceptable-content-type
          content-type
          (headers "accept"))
       (app req)
-      (http/error-response (i18n/tru "must accept {0}" content-type)
-                           http/status-not-acceptable))))
+      (http/error-response (tru "must accept {0}" content-type)
+                           HttpURLConnection/HTTP_NOT_ACCEPTABLE))))
+
+(defn verify-content-encoding
+  "Verification for the specified list of content-encodings."
+  [app allowed-encodings]
+  {:pre [(coll? allowed-encodings)
+         (every? string? allowed-encodings)]}
+  (fn [{:keys [headers request-method] :as req}]
+    (let [content-encoding (headers "content-encoding")]
+      (if (or (not= request-method :post)
+              (empty? content-encoding)
+              (some #{content-encoding} allowed-encodings))
+        (app req)
+        (http/error-response (tru "content encoding {0} not supported"
+                                  content-encoding)
+                             HttpURLConnection/HTTP_UNSUPPORTED_TYPE)))))
 
 (defn verify-content-type
   "Verification for the specified list of content-types."
@@ -124,36 +201,44 @@
     (if (= (:request-method req) :post)
       (let [content-type (headers "content-type")
             mediatype (if (nil? content-type) nil
-                          (str (media/base-type content-type)))]
+                          (kitchensink/base-type content-type))]
         (if (or (nil? mediatype) (some #{mediatype} content-types))
           (app req)
-          (http/error-response (i18n/tru "content type {0} not supported" mediatype)
-                               http/status-unsupported-type)))
+          (http/error-response (tru "content type {0} not supported" mediatype)
+                               HttpURLConnection/HTTP_UNSUPPORTED_TYPE)))
       (app req))))
 
-(defn validate-query-params
+(def params-schema {(s/optional-key :optional) [s/Str]
+                    (s/optional-key :required) [s/Str]})
+
+(defn-validated validate-query-params
   "Ring middleware that verifies that the query params in the request are legal
   based on the map `param-specs`, which contains a list of `:required` and
   `:optional` query parameters. If the validation fails, a 400 Bad Request is
   returned, with an explanation of the invalid parameters."
-  [app param-specs]
-  {:pre [(map? param-specs)
-         (= #{} (kitchensink/keyset (dissoc param-specs :required :optional)))
-         (every? string? (:required param-specs))
-         (every? string? (:optional param-specs))]}
+  [app param-specs :- params-schema]
   (fn [{:keys [params] :as req}]
-    (kitchensink/cond-let [p]
-                          (kitchensink/excludes-some params (:required param-specs))
-                          (http/error-response (i18n/tru "Missing required query parameter ''{0}''" p))
+    (if-let [excluded (kitchensink/excludes-some params (:required param-specs))]
+      (http/error-response (tru "Missing required query parameter ''{0}''"
+                                excluded))
+      (if-let [invalid (seq (set/difference (kitchensink/keyset params)
+                                            (set (:required param-specs))
+                                            (set (:optional param-specs))))]
+        (http/error-response (tru "Unsupported query parameter ''{0}''"
+                                  (first invalid)))
+        (app req)))))
 
-                          (let [diff (set/difference (kitchensink/keyset params)
-                                                     (set (:required param-specs))
-                                                     (set (:optional param-specs)))]
-                            (seq diff))
-                          (http/error-response (i18n/tru "Unsupported query parameter ''{0}''" (first p)))
-
-                          :else
-                          (app req))))
+(defn merge-param-specs
+  [& specs]
+  (letfn [(assoc-distinct-vals [result key & maps]
+            (cond-> result
+              (some key maps) (assoc key (distinct (mapcat key maps)))))]
+    (reduce (fn [result spec]
+              (-> (merge result spec)
+                  (assoc-distinct-vals :required result spec)
+                  (assoc-distinct-vals :optional result spec)))
+            nil
+            specs)))
 
 (defn validate-no-query-params
   "Ring middleware that verifies that there are no query params on the request.
@@ -250,32 +335,78 @@
   [app reject-large-commands? max-command-size]
   (fn [req]
     (if (= :post (:request-method req))
-      (let [length-in-bytes (request/content-length req)]
+      (let [length-in-bytes (or (when-let [length (get-in req [:headers "x-uncompressed-length"])]
+                                  (try
+                                    (Long/parseLong length)
+                                    (catch NumberFormatException _
+                                      (log/warn (trs "The X-Uncompressed-Length value {0} cannot be converted to a long."
+                                                     length)))))
+                                (request/content-length req))]
 
-        (when length-in-bytes
-          (log/debug (i18n/trs "Processing command with a content-length of {0} bytes" length-in-bytes))
-          (let [{:strs [command version]} (:query-params req)]
-            (update! (cmd/global-metric :size) length-in-bytes)
-            (when (and command version)
-              (cmd/create-metrics-for-command! command version)
-              (update! (cmd/cmd-metric command version :size) length-in-bytes))))
-        (when-not length-in-bytes
-          (log/warn (i18n/trs "No content length found for POST. POST bodies that are too large could cause memory-related failures.")))
+        (let [{:strs [command version]} (:query-params req)]
+          (if length-in-bytes
+            (do (log/debug (trs "Processing command with a content-length of {0} bytes" length-in-bytes))
+                (update! (cmd/global-metric :size) length-in-bytes)
+                (when (and command version)
+                  (let [command (const/normalize-command-name command)]
+                    ;; command name must be normalized so correct metrics are updated
+                    (cmd/create-metrics-for-command! command version)
+                    (update! (cmd/cmd-metric command version :size) length-in-bytes))))
+            (log/warn (trs "Neither Content-Length or X-Uncompressed-Length header is set.
+                            This {0} command will not be counted in command size metrics"
+                           command))))
 
         (if (and length-in-bytes
                  reject-large-commands?
                  (> length-in-bytes max-command-size))
           (do
-            (log/warn (i18n/trs "content-length of command is {0} bytes and is larger than the maximum allowed command size of {1} bytes"
-                                length-in-bytes
-                                max-command-size))
+            (log/warn (trs "content-length of command is {0} bytes and is larger than the maximum allowed command size of {1} bytes"
+                           length-in-bytes
+                           max-command-size))
             (consume-and-close (:body req) length-in-bytes)
             (http/error-response
-             (i18n/tru "Command rejected due to size exceeding max-command-size")
-             http/status-entity-too-large))
+             (tru "Command rejected due to size exceeding max-command-size")
+             HttpURLConnection/HTTP_ENTITY_TOO_LARGE))
           (app req)))
       (app req))))
 
+;; for testing via with-redefs
+(defn get-sync-ver []
+  constants/pdb-sync-ver)
+
+(defn verify-sync-version
+  "Check that the x-pdb-sync-ver header of an incoming requests matches the
+   pdb-sync-ver locally. This header is only sent from pe-puppetdb sync requests
+   and a mismatch indicates that one pdb has upgraded in an incompatible way
+   before the other. In this case all sync requests will return 409s until both
+   sides have been upgraded and have matching sync versions. If the header is
+   present but improperly formatted returns a 400 error response."
+  [app]
+  (fn [{{:strs [x-pdb-sync-ver]} :headers :as req}]
+    (let [maybe-sync-ver (try
+                           (some-> x-pdb-sync-ver Integer/parseInt)
+                           (catch NumberFormatException _
+                             {:error true :input x-pdb-sync-ver}))]
+      (cond
+        (nil? maybe-sync-ver) (app req)
+
+        (integer? maybe-sync-ver)
+        (if (= (get-sync-ver) maybe-sync-ver)
+          (app req)
+          (http/error-response
+           (tru "PDB sync request version {0} too new for this server (expected {1})."
+                maybe-sync-ver (get-sync-ver))
+           409))
+
+        (:error maybe-sync-ver)
+        (http/error-response
+         (tru "The x-pdb-sync-ver header: {0} cannot be converted to an int."
+              (:input maybe-sync-ver)))
+
+        :else (http/error-response
+               (tru "Unknown sync version check state")
+               500)))))
+
 (defn wrap-with-puppetdb-middleware
   "Default middleware for puppetdb webservers."
   [app]
@@ -290,9 +421,9 @@
   "Middleware that checks the parent exists before serving the rest of the
    application. This ensures we always return 404's on child paths when the
    parent data is empty."
-  [app version parent route-param-key]
+  [app _version parent route-param-key]
   (fn [{:keys [globals route-params] :as req}]
-    (let [{:keys [scf-read-db url-prefix]} globals]
+    (let [{:keys [scf-read-db]} globals]
       ;; There is a race condition here, in particular we open up 1 transaction
       ;; for the parent test, but the rest of the query results are done via the
       ;; streaming query. This can't be solved until we work out a way to
@@ -300,13 +431,16 @@
       (if (jdbc/with-transacted-connection scf-read-db
             (qe/object-exists? parent (get route-params route-param-key)))
         (app req)
-        (http/json-response {:error (i18n/tru "No information is known about {0} {1}" (name parent) (get route-params route-param-key))} http/status-not-found)))))
+        (http/json-response {:error (tru "No information is known about {0} {1}"
+                                         (name parent)
+                                         (get route-params route-param-key))}
+                            HttpURLConnection/HTTP_NOT_FOUND)))))
 
-(pls/defn-validated url-decode :- s/Str
+(defn-validated url-decode :- s/Str
   [x :- s/Str]
-  (java.net.URLDecoder/decode x))
+  (java.net.URLDecoder/decode x "utf-8"))
 
-(pls/defn-validated make-pdb-handler :- handler-schema
+(defn-validated make-pdb-handler :- handler-schema
   "Similar to `bidi.ring/make-handler` but does not merge route-params
   into the regular parameter map. Currently route-params causes
   validation errors with merged in with parameters. Parameter names
@@ -324,4 +458,3 @@
           (handler-fn handler)
           (update req :route-params merge (kitchensink/mapvals url-decode route-params))
           (apply dissoc match-context :handler (keys req))))))))
-
