@@ -62,21 +62,17 @@
    applying ordering constraints."
   (:require [clojure.string :as str]
             [puppetlabs.i18n.core :as i18n]
-            [puppetlabs.kitchensink.core :as kitchensink]
-            [puppetlabs.puppetdb.jdbc :as jdbc]
-            [puppetlabs.puppetdb.honeysql :as h]
+            [puppetlabs.kitchensink.core :refer [parse-number keyset]]
+            [puppetlabs.puppetdb.query.common :refer [bad-query-ex]]
             [puppetlabs.puppetdb.utils :as utils]
+            [puppetlabs.puppetdb.utils.string-formatter :as formatter]
             [puppetlabs.puppetdb.time :refer [to-timestamp]]
-            [puppetlabs.kitchensink.core :refer [parse-number keyset valset order-by-expr?]]
             [puppetlabs.puppetdb.scf.storage-utils :as sutils
              :refer [db-serialize sql-as-numeric sql-array-query-string
-                     legacy-sql-regexp-match sql-regexp-array-match]]
-            [puppetlabs.puppetdb.jdbc
-             :refer [valid-jdbc-query? limited-query-to-vec query-to-vec paged-sql count-sql get-result-count]]
-            [puppetlabs.puppetdb.query.paging :refer [requires-paging?]]
+                     legacy-sql-regexp-match sql-regexp-array-match-str]]
+            [puppetlabs.puppetdb.jdbc :refer [valid-jdbc-query?]]
             [clojure.core.match :refer [match]]
-            [schema.core :as s]
-            [puppetlabs.puppetdb.utils :as utils]))
+            [schema.core :as s]))
 
 (defn wrap-with-supported-fns
   [schema]
@@ -85,66 +81,8 @@
           (s/optional-key :min) s/Any
           (s/optional-key :max) s/Any
           (s/optional-key :avg) s/Any
-          (s/optional-key :sum) s/Any}))
-
-(defn execute-paged-query*
-  "Helper function to executed paged queries.  Builds up the paged sql string,
-  executes the query, and returns map containing the `:result` key and an
-  optional `:count` key."
-  [fail-limit query {:keys [limit offset order_by include_total] :as paging-options}]
-  {:pre [(and (integer? fail-limit) (>= fail-limit 0))
-         (valid-jdbc-query? query)
-         ((some-fn nil? integer?) limit)
-         ((some-fn nil? integer?) offset)
-         ((some-fn nil? sequential?) order_by)
-         (every? order-by-expr? order_by)]
-   :post [(map? %)
-          (vector? (:result %))
-          ((some-fn nil? integer?) (:count %))]}
-  (let [[sql & params] (if (string? query) [query] query)
-        paged-sql      (paged-sql sql paging-options)
-        result         {:result
-                        (limited-query-to-vec
-                         fail-limit
-                         (apply vector paged-sql params))}]
-    ;; TODO: this could also be implemented using `COUNT(*) OVER()`,
-    ;; which would allow us to get the results and the count via a
-    ;; single query (rather than two separate ones).  Need to do
-    ;; some benchmarking to see which is faster.
-    (if include_total
-      (assoc result :count
-             (get-result-count (apply vector (count-sql sql) params)))
-      result)))
-
-(defn execute-query*
-  "Helper function to executed non-paged queries.  Returns a map containing the
-  `:result` key."
-  [fail-limit query]
-  {:pre [(and (integer? fail-limit) (>= fail-limit 0))
-         (valid-jdbc-query? query)]
-   :post [(map? %)
-          (vector? (:result %))]}
-  {:result (limited-query-to-vec fail-limit query)})
-
-(defn execute-query
-  "Given a query and a map of paging options, adds the necessary SQL for
-  implementing the paging, executes the query, and returns a map containing
-  the results and metadata.
-
-  The return value will contain a key `:result`, whose value is a vector of
-  the query results.  If the paging options indicate that a 'total record
-  count' should be returned, then the map will also include a key `:count`,
-  whose value is an integer indicating the total number of results available."
-  ([query paging-options] (execute-query 0 query paging-options))
-  ([fail-limit query {:keys [limit offset order_by] :as paging-options}]
-     {:pre [((some-fn string? sequential?) query)]
-      :post [(map? %)
-             (vector? (:result %))
-             ((some-fn nil? integer?) (:count %))]}
-     (let [sql-and-params (if (string? query) [query] query)]
-       (if (requires-paging? paging-options)
-         (execute-paged-query* fail-limit sql-and-params paging-options)
-         (execute-query* fail-limit sql-and-params)))))
+          (s/optional-key :sum) s/Any
+          (s/optional-key :to_string) s/Str}))
 
 (defn compile-term
   "Compile a single query term, using `ops` as the set of legal operators. This
@@ -157,14 +95,14 @@
     {:where nil :params nil}
 
     (not op)
-    (throw (IllegalArgumentException.
+    (throw (bad-query-ex
             (i18n/tru "{0} is not well-formed: queries must contain at least one operator"
                       (vec term))))
 
     :else
     (if-let [f (ops op)]
       (apply f args)
-      (throw (IllegalArgumentException.
+      (throw (bad-query-ex
               (i18n/tru "{0} is not well-formed: query operator ''{1}'' is unknown"
                         (vec term) op))))))
 
@@ -177,7 +115,7 @@
   {:pre  [(every? coll? terms)]
    :post [(string? (:where %))]}
   (when (empty? terms)
-    (throw (IllegalArgumentException.
+    (throw (bad-query-ex
             (i18n/tru "{0} requires at least one term" op))))
   (let [compiled-terms (map #(compile-term ops %) terms)
         params (mapcat :params compiled-terms)
@@ -206,10 +144,10 @@
   "Compile a NOT operator, applied to `term`. This term simply negates the
   value of `term`. Basically this function just serves as error checking for
   `negate-term*`."
-  [version ops & terms]
+  [_version ops & terms]
   {:post [(string? (:where %))]}
   (when-not (= (count terms) 1)
-    (throw (IllegalArgumentException.
+    (throw (bad-query-ex
             (i18n/tru "''not'' takes exactly one argument, but {0} were supplied" (count terms)))))
   (negate-term* ops (first terms)))
 
@@ -260,6 +198,7 @@
    "line"                   ["resource_events"]
    "containment_path"       ["resource_events"]
    "containing_class"       ["resource_events"]
+   "name"                   ["resource_events"]
    "environment"            ["environments"]})
 
 (def resource-event-columns
@@ -275,6 +214,7 @@
    "resource_type"          ["latest_events"]
    "resource_title"         ["latest_events"]
    "property"               ["latest_events"]
+   "name"                   ["latest_events"]
    "new_value"              ["latest_events"]
    "old_value"              ["latest_events"]
    "message"                ["latest_events"]
@@ -303,6 +243,7 @@
    "line"                   ["resource_events"]
    "containment_path"       ["resource_events"]
    "containing_class"       ["resource_events"]
+   "name"                   ["resource_events"]
    "environment"            ["environments"]})
 
 (def report-columns
@@ -339,10 +280,10 @@
   "This function takes a query type (:resource, :fact, :node) and a query
    API version number, and returns a set of strings which are the names the
    fields that are legal to query"
-  (fn [query-type query-api-version] query-type))
+  (fn [query-type _query-api-version] query-type))
 
 (defmethod queryable-fields :resource
-  [_ query-api-version]
+  [_ _query-api-version]
   (keyset resource-columns))
 
 (defmethod queryable-fields :fact
@@ -372,11 +313,11 @@
   (let [[subselect & params] (compile-term ops subquery)
         subquery-type (subquery->type (first subquery))]
     (when-not subquery-type
-      (throw (IllegalArgumentException.
+      (throw (bad-query-ex
               (i18n/tru "The argument to extract must be a select operator, not ''{0}''"
                         (first subquery)))))
     (when-not (get (queryable-fields subquery-type query-api-version) field)
-      (throw (IllegalArgumentException.
+      (throw (bad-query-ex
               (i18n/tru "Can't extract unknown {0} field ''{1}''. Acceptable fields are: {2}"
                         (name subquery-type) field (str/join ", " (sort (queryable-fields subquery-type query-api-version)))))))
     {:where (format "SELECT r1.%s FROM (%s) r1" field subselect)
@@ -392,11 +333,11 @@
    :post [(map? %)
           (string? (:where %))]}
   (when-not (get (queryable-fields kind query-api-version) field)
-    (throw (IllegalArgumentException.
+    (throw (bad-query-ex
              (i18n/tru "Can''t match on unknown {0} field ''{1}'' for ''in''. Acceptable fields are: {2}"
                        (name kind) field (str/join ", " (sort (queryable-fields kind query-api-version)))))))
   (when-not (= (first subquery) "extract")
-    (throw (IllegalArgumentException.
+    (throw (bad-query-ex
             (i18n/tru "The subquery argument of ''in'' must be an ''extract'', not ''{0}''"
                       (first subquery)))))
   (let [{:keys [where] :as compiled-subquery} (compile-term ops subquery)
@@ -437,23 +378,27 @@
   {:post [(valid-jdbc-query? %)]}
   (let [{:keys [where params]} (compile-term ops query)
         sql (format "SELECT %s FROM (
-                      SELECT fs.certname,
-                             fp.name as name,
-                             fv.value,
-                             env.environment
-                      FROM factsets fs
-                        INNER JOIN facts as f on fs.id = f.factset_id
-                        INNER JOIN fact_values as fv on f.fact_value_id = fv.id
-                        INNER JOIN fact_paths as fp on f.fact_path_id = fp.id
-                        LEFT OUTER JOIN environments as env on fs.environment_id = env.id
-                      WHERE depth = 0) AS facts
+                     SELECT certname,
+                            environment,
+                            value #>> '{}' AS value,
+                            producer_timestamp,
+                            key AS name
+                     FROM (SELECT certname,
+                                  producer_timestamp,
+                                  environment_id,
+                                  facts.*
+                           FROM factsets,
+                           lateral (select * from jsonb_each(stable)
+                                    union all select * from jsonb_each(volatile)) as facts) fs
+                           LEFT JOIN environments env ON fs.environment_id = env.id
+                     ) AS facts
                     WHERE %s" (column-map->sql fact-columns) where)]
     (apply vector sql params)))
 
 (defn certname-names-query [active]
   (if active
-    "SELECT name FROM certnames WHERE deactivated IS NULL AND expired IS NULL"
-    "SELECT name FROM certnames WHERE deactivated IS NOT NULL OR expired IS NOT NULL") )
+    "SELECT certname FROM certnames_status WHERE deactivated IS NULL AND expired IS NULL"
+    "SELECT certname FROM certnames_status WHERE deactivated IS NOT NULL OR expired IS NOT NULL"))
 
 (defn compile-resource-equality
   "Compile an = operator for a resource query. `path` represents the field
@@ -462,7 +407,7 @@
   {:post [(map? %)
           (:where %)]}
   (when-not (= (count args) 2)
-    (throw (IllegalArgumentException.
+    (throw (bad-query-ex
             (i18n/tru "= requires exactly two arguments, but {0} were supplied"
                       (count args)))))
   (match [path]
@@ -497,7 +442,7 @@
           :params [value]}
 
          ;; ...else, failure
-         :else (throw (IllegalArgumentException.
+         :else (throw (bad-query-ex
                        (i18n/tru "''{0}'' is not a queryable object for resources in the version {1} API"
                                  path (last (name version)))))))
 
@@ -505,13 +450,12 @@
   "Compile an '~' predicate for a resource query, which does regexp matching.
   This is done by leveraging the correct database-specific regexp syntax to
   return only rows where the supplied `path` match the given `pattern`."
-  [version & [path value :as args]]
+  [version path value]
   {:post [(map? %)
           (:where %)]}
   (match [path]
          ["tag"]
-         {:where (h/sqlraw->str
-                   (sql-regexp-array-match "tags"))
+         {:where (sql-regexp-array-match-str "tags")
           :params [value]}
 
          ;; node join.
@@ -529,7 +473,7 @@
           :params [value]}
 
          ;; ...else, failure
-         :else (throw (IllegalArgumentException.
+         :else (throw (bad-query-ex
                        (i18n/tru "''{0}'' cannot be the target of a regexp match for version {1} of the resources API"
                                  path (last (name version)))))))
 
@@ -562,7 +506,7 @@
            {:where (format "facts.certname IN (%s)" (certname-names-query value))}
 
            :else
-           (throw (IllegalArgumentException.
+           (throw (bad-query-ex
                    (i18n/tru "{0} is not a queryable object for version {1} of the facts query api"
                              path (last (name version))))))))
 
@@ -593,7 +537,7 @@
                              (legacy-sql-regexp-match "facts.value"))
               :params [pattern]}
 
-             :else (throw (IllegalArgumentException.
+             :else (throw (bad-query-ex
                            (i18n/tru "{0} is not a valid version {1} operand for regexp comparison"
                                      path (last (name version)))))))))
 
@@ -613,9 +557,9 @@
            {:where  (format "%s %s ? and depth = 0" (sql-as-numeric "facts.value") op)
             :params [number]}
 
-           :else (throw (IllegalArgumentException.
+           :else (throw (bad-query-ex
                          (i18n/tru "{0} is not a queryable object for facts" path))))
-    (throw (IllegalArgumentException.
+    (throw (bad-query-ex
             (i18n/tru "Value {0} must be a number for {1} comparison." value op)))))
 
 (defn compile-resource-event-inequality
@@ -626,7 +570,7 @@
   {:post [(map? %)
           (string? (:where %))]}
   (when-not (= (count args) 3)
-    (throw (IllegalArgumentException.
+    (throw (bad-query-ex
             (i18n/tru "{0} requires exactly two arguments, but {1} were supplied"
                       op (dec (count args))))))
 
@@ -635,14 +579,14 @@
                           "run_end_time"        "reports.end_time"
                           "report_receive_time" "reports.receive_time"}]
     (match [path]
-           [(field :guard (kitchensink/keyset timestamp-fields))]
+           [(field :guard (keyset timestamp-fields))]
            (if-let [timestamp (to-timestamp value)]
              {:where (format "%s %s ?" (timestamp-fields field) op)
-              :params [(to-timestamp value)]}
-             (throw (IllegalArgumentException.
+              :params [timestamp]}
+             (throw (bad-query-ex
                      (i18n/tru "''{0}'' is not a valid timestamp value" value))))
 
-           :else (throw (IllegalArgumentException.
+           :else (throw (bad-query-ex
                          (i18n/tru "{0} operator does not support object ''{1}'' for resource events"
                                    op path))))))
 
@@ -654,10 +598,10 @@
     {:post [(map? %)
             (string? (:where %))]}
     (when-not (= (count args) 2)
-      (throw (IllegalArgumentException.
+      (throw (bad-query-ex
               (i18n/tru "= requires exactly two arguments, but {0} were supplied"
                         (count args)))))
-    (let [path (utils/dashes->underscores path)]
+    (let [path (formatter/dashes->underscores path)]
       (match [path]
              ["certname"]
              {:where "reports.certname = ?"
@@ -668,7 +612,7 @@
               :params [value]}
 
              ["latest_report?"]
-             (let [latest-report-clause (format "resource_events.report_id %s (SELECT certnames.latest_report_id FROM certnames)"
+             (let [latest-report-clause (format "resource_events.report_id %s (SELECT reports_latest.id FROM reports_latest)"
                                                 (if value "IN" "NOT IN"))]
                {:where latest-report-clause
                 :latest-report-clause latest-report-clause})
@@ -697,7 +641,7 @@
              {:where (format "resource_events.%s = ? AND resource_events.%s IS NOT NULL" field field)
               :params [(db-serialize value)] }
 
-             :else (throw (IllegalArgumentException.
+             :else (throw (bad-query-ex
                            (i18n/tru "''{0}'' is not a queryable object for version {1} of the resource events API"
                                      path (last (name version)))))))))
 
@@ -709,9 +653,9 @@
     {:post [(map? %)
             (string? (:where %))]}
     (when-not (= (count args) 2)
-      (throw (IllegalArgumentException.
+      (throw (bad-query-ex
                (i18n/tru "~ requires exactly two arguments, but {0} were supplied" (count args)))))
-    (let [path (utils/dashes->underscores path)]
+    (let [path (formatter/dashes->underscores path)]
       (match [path]
              ["certname"]
              {:where (legacy-sql-regexp-match "reports.certname")
@@ -734,7 +678,7 @@
                              field)
               :params [pattern]}
 
-             :else (throw (IllegalArgumentException.
+             :else (throw (bad-query-ex
                            (i18n/tru "''{0}'' is not a queryable object for version {1} of the resource events API"
                                      path (last (name version)))))))))
 
@@ -745,15 +689,15 @@
   {:post [(map? %)
           (string? (:where %))]}
   (when-not (= (count args) 2)
-    (throw (IllegalArgumentException.
+    (throw (bad-query-ex
             (i18n/tru "= requires exactly two arguments, but {0} were supplied" (count args)))))
-  (let [db-field (utils/dashes->underscores path)]
+  (let [db-field (formatter/dashes->underscores path)]
     (match [db-field]
            [(field :guard fields)]
            {:where (format "%s = ?" field)
             :params [value]}
 
-           :else (throw (IllegalArgumentException.
+           :else (throw (bad-query-ex
                          (i18n/tru "{0} is not a queryable object for event counts" path))))))
 
 (defn compile-event-count-inequality
@@ -763,7 +707,7 @@
   {:post [(map? %)
           (string? (:where %))]}
   (when-not (= (count args) 3)
-    (throw (IllegalArgumentException.
+    (throw (bad-query-ex
             (i18n/tru "{0} requires exactly two arguments, but {1} were supplied"
                       op (dec (count args))))))
   (match [path]
@@ -771,7 +715,7 @@
          {:where (format "%s %s ?" field op)
           :params [value]}
 
-         :else (throw (IllegalArgumentException.
+         :else (throw (bad-query-ex
                        (i18n/tru "{0} operator does not support object ''{1}'' for event counts"
                                  op path)))))
 
